from moai.data.datasets.human.genebody import GeneBody
from moai.data.datasets.human.amass import AMASS
from moai.data.datasets.human.fit3d import Fit3D
from moai.data.datasets.human.thuman2 import THuman2

__all__ = [
    'GeneBody',
    'AMASS',
    'Fit3D',
    'THuman2',
]